from heaserver.service.runner import routes
from heaserver.service.db import awsservicelib
from heaserver.service import response
from heaserver.service.appproperty import HEA_DB
from aiohttp import web


@routes.get('/accounts/{id}')
async def get_volume_id(request: web.Request) -> web.Response:
    volume_id = await awsservicelib.get_volume_id_for_account_id(request)
    if volume_id is not None:
        return response.status_ok(body=volume_id)
    else:
        return response.status_not_found()


@routes.get('/properties/{id}')
async def get_properties_id(request: web.Request) -> web.Response:
    property = await request.app[HEA_DB].get_property(request.app, request.match_info['id'])
    if property is not None:
        return response.status_ok(body=property.to_json())
    else:
        return response.status_not_found()


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3folders/{id}/duplicator')
async def post_folder_duplicator(request: web.Request) -> web.Response:
    return await awsservicelib.copy_object(request)


@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3folders/')
@routes.post('/volumes/{volume_id}/buckets/{bucket_id}/awss3folders')
async def post_folder(request: web.Request) -> web.Response:
    return await awsservicelib.create_object(request)



"""
Creates a test case class for use with the unittest library that is built into Python.
"""
from heaobject.root import Permission

from . import service
from heaserver.service.testcase.microservicetestcase import get_test_case_cls_default
from heaobject.user import NONE_USER, TEST_USER, ALL_USERS
from heaserver.service.testcase.expectedvalues import Action, Link
from datetime import datetime, timezone

db_store = {
    service.MONGODB_ORGANIZATION_COLLECTION: [
        {
            "id": "666f6f2d6261722d71757578",
            "instance_id": "heaobject.organization.Organization^666f6f2d6261722d71757578",
            "source": None,
            "source_detail": None,
            "name": "Bob",
            "display_name": "Bob",
            "description": "Description of Bob",
            "owner": NONE_USER,
            "created": datetime(2022, 5, 17, 0, 0, 0, 0, tzinfo=timezone.utc),
            "modified": datetime(2022, 5, 17, 0, 0, 0, 0, tzinfo=timezone.utc),
            "invites": [],
            "shares": [{
                'type': 'heaobject.root.ShareImpl',
                'invite': None,
                'user': ALL_USERS,
                'permissions': [Permission.CHECK_DYNAMIC.name],
                'type_display_name': 'Share'
            }],
            "derived_by": None,
            "derived_from": [],
            "account_ids": [],
            "principal_investigator_id": "23423DAFSDF12adfasdf3",
            "manager_ids": [],
            "member_ids": [],
            'type': 'heaobject.organization.Organization',
            'mime_type': 'application/x.organization',
            'admin_ids': [TEST_USER],
            'type_display_name': 'Organization',
            'admin_group_ids': [],
            'manager_group_ids': [],
            'member_group_ids': [],
            'collaborator_ids': []
        },
        {
            "id": "0123456789ab0123456789ab",
            "instance_id": "heaobject.organization.Organization^0123456789ab0123456789ab",
            "source": None,
            "source_detail": None,
            "name": "Reximus",
            "display_name": "Reximus",
            "description": "Description of Reximus",
            "owner": NONE_USER,
            "created": datetime(2022, 5, 17, 0, 0, 0, 0, tzinfo=timezone.utc),
            "modified": datetime(2022, 5, 17, 0, 0, 0, 0, tzinfo=timezone.utc),
            "invites": [],
            "shares": [{
                'type': 'heaobject.root.ShareImpl',
                'invite': None,
                'user': ALL_USERS,
                'permissions': [Permission.CHECK_DYNAMIC.name],
                'type_display_name': 'Share'
            }],
            "derived_by": None,
            "derived_from": [],
            "account_ids": [],
            "principal_investigator_id": "11234867890b0123a56789ab",
            "manager_ids": [],
            "member_ids": [],
            'type': 'heaobject.organization.Organization',
            'mime_type': 'application/x.organization',
            'admin_ids': [TEST_USER],
            'type_display_name': 'Organization',
            'admin_group_ids': [],
            'manager_group_ids': [],
            'member_group_ids': [],
            'collaborator_ids': []
        }
    ]}

content = {
    service.MONGODB_ORGANIZATION_COLLECTION: {
        '666f6f2d6261722d71757578': b'The quick brown fox jumps over the lazy dog'
    }
}


OrganizationTestCase = get_test_case_cls_default(coll=service.MONGODB_ORGANIZATION_COLLECTION,
                                                 wstl_package=service.__package__,
                                                 href='http://localhost:8080/organizations/',
                                                 fixtures=db_store,
                                                 content=content,
                                                 get_actions=[Action(name='heaserver-organizations-organization-get-properties',
                                                                         rel=['properties']),
                                                              Action(name='heaserver-organizations-organization-get-open-choices',
                                                                         url='http://localhost:8080/organizations/{id}/opener',
                                                                         rel=['hea-opener-choices']),
                                                              Action(name='heaserver-organizations-organization-duplicate',
                                                                         url='http://localhost:8080/organizations/{id}/duplicator',
                                                                         rel=['duplicator'])
                                                              ],
                                                 get_all_actions=[Action(name='heaserver-organizations-organization-get-properties',
                                                                             rel=['properties']),
                                                                  Action(name='heaserver-organizations-organization-get-open-choices',
                                                                             url='http://localhost:8080/organizations/{id}/opener',
                                                                             rel=['hea-opener-choices']),
                                                                  Action(name='heaserver-organizations-organization-duplicate',
                                                                             url='http://localhost:8080/organizations/{id}/duplicator',
                                                                             rel=['duplicator'])],
                                                 expected_opener=Link(url=f'http://localhost:8080/organizations/{db_store[service.MONGODB_ORGANIZATION_COLLECTION][0]["id"]}/content', rel=['hea-default', 'hea-opener', 'text/plain']),
                                                 duplicate_action_name='heaserver-organizations-organization-duplicate-form',
                                                 put_content_status=204, sub=TEST_USER)



"""
Functions for interacting with Amazon Web Services.

This module supports management of AWS accounts, S3 buckets, and objects in S3 buckets. It uses Amazon's boto3 library
behind the scenes.

In order for HEA to access AWS accounts, buckets, and objects, there must be a volume accessible to the user through
the volumes microservice with an AWSFileSystem for its file system. Additionally, credentials must either be stored
in the keychain microservice and associated with the volume through the volume's credential_id attribute,
or stored on the server's file system in a location searched by the AWS boto3 library. Users can only see the
accounts, buckets, and objects to which the provided AWS credentials allow access, and HEA may additionally restrict
the returned objects as documented in the functions below. The purpose of volumes in this case is to supply credentials
to AWS service calls. Support for boto3's built-in file system search for credentials is only provided for testing and
should not be used in a production setting. This module is designed to pass the current user's credentials to AWS3, not
to have application-wide credentials that everyone uses.

The request argument to these functions is expected to have a OIDC_CLAIM_sub header containing the user id for
permissions checking. No results will be returned if this header is not provided or is empty.

In general, there are two flavors of functions for getting accounts, buckets, and objects. The first expects the id
of a volume as described above. The second expects the id of an account, bucket, or bucket and object. The latter
attempts to match the request up to any volumes with an AWSFileSystem that the user has access to for the purpose of
determine what AWS credentials to use. They perform the
same except when the user has access to multiple such volumes, in which case supplying the volume id avoids a search
through the user's volumes.
"""
from aiohttp import web
from heaobject.awss3key import display_name
from .. import response
from ..appproperty import HEA_DB
from aiohttp.web import HTTPError
from heaobject.volume import AWSFileSystem
from heaserver.service import aiohttp
from asyncio import gather
from botocore.exceptions import ClientError as BotoClientError
from .aws import client_error_status


async def get_volume_id_for_account_id(request: web.Request) -> str | None:
    """
    Gets the id of the volume associated with an AWS account. The account id is expected to be in the request object's
    match_info mapping, with key 'id'. The account is fetched from AWS during this operation.

    :param request: an aiohttp Request object (required).
    :return: a volume id string, or None if no volume was found associated with the AWS account or the account does not
    exist.
    """
    async def get_one(request, volume_id):
        return volume_id, await request.app[HEA_DB].get_account(request, volume_id)
    return next((volume_id for (volume_id, a) in await gather(
        *[get_one(request, v.id) async for v in request.app[HEA_DB].get_volumes(request, AWSFileSystem)])
                 if
                 a.id == request.match_info['id']), None)


def handle_client_error(e: BotoClientError) -> HTTPError:
    """
    Translates a boto3 client error into an appropriate HTTP response.

    :param e: a boto3 client error (required).
    :return: an HTTP response with status code >=400 that can be raised as an exception.
    """
    status, msg = client_error_status(e)
    return response.status_generic_error(status=status, body=msg)


def s3_object_display_name(bucket_name: str, key: str | None) -> str:
    """
    Function for use with heaserver.service.aiohttp.http_error_message() to generate names of S3 buckets and objects
    for use in error messages.

    :param bucket_name: the name of an S3 bucket (required).
    :param key: an optional key for objects in a bucket.
    """
    return f'{display_name(key)} in bucket {bucket_name}' if key is not None else bucket_name


def http_error_message(http_error: web.HTTPError, bucket_name: str, key: str | None) -> web.HTTPError:
    """
    If the HTTPError object has an empty body, it will try filling in the body with a message appropriate for the given
    status code for operations on desktop objects from AWS. Uses s3_object_display_name() to generate the message.

    :param http_error: the HTTPError (required).
    :param bucket_name: the bucket name (required).
    :param key: the key.
    :return: the updated HTTPError.
    """
    return aiohttp.http_error_message(http_error, s3_object_display_name, bucket_name, key)


def _activity_object_display_name(bucket_name: str, key: str | None) -> str:
    return s3_object_display_name(bucket_name, key)
activity_object_display_name = _activity_object_display_name




import configparser
from contextlib import AbstractContextManager, ExitStack
from typing import Optional, Sequence, Mapping, IO

from aiohttp import web
from heaobject.root import DesktopObjectDict, DesktopObject

from heaserver.service.db.database import InMemoryDatabase, CollectionKey, query_content, query_fixtures
from heaserver.service.db.mongo import Mongo
from heaserver.service.db.opensearch import OpenSearch, OpenSearchManager, ItemTypeVar, MongoWithOpenSearch, \
    S3WithOpenSearch
from heaserver.service.testcase.mockdatabase import MockDatabase
from .util import freeze_time


class MockOpenSearch(MockDatabase, InMemoryDatabase, OpenSearch ):
    """
    Mock implementation of the Mongo class.

    It does not implement an aggregate method due the lack of good ways to mock such a thing.
    """

    def __init__(self, config: Optional[configparser.ConfigParser] = None,
                 opensearch: Optional['MockOpenSearch'] = None,
                 **kwargs) -> None:
        """
        Sets the db property of the app context with a motor MongoDB client instance.

        :param config: a configparser.ConfigParser object. MockMongo does not have a config section of its own.
        """
        super().__init__(config, **kwargs)
        if opensearch is not None:
            self.add_desktop_objects(opensearch.get_all_desktop_objects())
            self.add_content(opensearch.get_all_content())

    async def search(self,query: dict, search_item_type: type[ItemTypeVar],
                     index: str | None = None,
                     max_results: int = 1000,
                     size: int = 100) -> list[ItemTypeVar]:
        """
        Gets an object from the index.

        :param request: the aiohttp Request object (required).
        used as the attributes to query by.
        :param search_item_type: The class to be instantiated for holding search data
        :param index: name of opensearch index
        :param max_results: maximum number of results to return
        :param size: number of results to return
        :return: a list of SearchItems
        """

        return [search_item_type()]


    async def ping(self):
        """
        Raises an exception if the database does not respond to a ping command. The mock implementation never throws
        an exception.
        """
        pass


class MockOpenSearchManager(OpenSearchManager):
    """
    Database manager for a mock of OpeanSearch that stores collections in memory.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__opensearch: Optional[MockOpenSearch] = None

    @classmethod
    def get_context(cls) -> list[AbstractContextManager]:
        result = super().get_context()
        result.append(freeze_time())
        return result

    def start_database(self, context_manager: ExitStack) -> None:
        self.__opensearch = MockOpenSearch()
        super().start_database(context_manager)

    def insert_desktop_objects(self, desktop_objects: Optional[Mapping[CollectionKey, Sequence[DesktopObjectDict]]]):
        super().insert_desktop_objects(desktop_objects)
        assert self.__opensearch is not None
        if desktop_objects:
            self.__opensearch.add_desktop_objects({k: v for k, v in query_fixtures(desktop_objects, db_manager=MockOpenSearchManager).items() if k is not None})

    def insert_content(self, content: Optional[Mapping[CollectionKey, Mapping[str, bytes]]]):
        super().insert_content(content)
        assert self.__opensearch is not None
        if content:
            self.__opensearch.add_content({k: v for k, v in query_content(content, db_manager=MockOpenSearchManager).items() if k is not None})

    def get_database(self) -> MockOpenSearch:
        """
        Gets a mock mongo database object (only available after start_database() has been called).
        """
        if self.__opensearch is None:
            raise ValueError('start_database not called')
        return self.__opensearch

    def get_mongo(self) -> MockOpenSearch | None:
        return self.__opensearch


class MockMongoWithOpenSearch(MockDatabase, InMemoryDatabase, MongoWithOpenSearch):
    """
    Mock implementation of the MongoWithOpenSearch class.
    """
    def __init__(self, config: Optional[configparser.ConfigParser] = None,
                 mongo_opensearch: Optional['MockMongoWithOpenSearch'] = None,
                 **kwargs) -> None:
        super().__init__(config, **kwargs)
        if mongo_opensearch is not None:
            self.add_desktop_objects(mongo_opensearch.get_all_desktop_objects())
            self.add_content(mongo_opensearch.get_all_content())

    async def search(self, query: dict,  search_item_type: type[ItemTypeVar],
                     index: str | None = None,
                     max_results: int = 1000,
                     size: int = 100) -> list[ItemTypeVar]:
        return [search_item_type()]

    async def ping(self):
        pass


class MockMongoWithOpenSearchManager(OpenSearchManager):
    """
    Mock manager for MongoWithOpenSearch that stores collections in memory.
    """
    def __init__(self) -> None:
        super().__init__()
        self.__mongo_opensearch: Optional[MockMongoWithOpenSearch] = None

    @classmethod
    def get_context(cls) -> list[AbstractContextManager]:
        result = super().get_context()
        result.append(freeze_time())
        return result

    def start_database(self, context_manager: ExitStack) -> None:
        self.__mongo_opensearch = MockMongoWithOpenSearch()
        super().start_database(context_manager)

    def insert_desktop_objects(self, desktop_objects: Optional[Mapping[CollectionKey, Sequence[DesktopObjectDict]]]):
        super().insert_desktop_objects(desktop_objects)
        assert self.__mongo_opensearch is not None
        if desktop_objects:
            self.__mongo_opensearch.add_desktop_objects({k: v for k, v in query_fixtures(desktop_objects, db_manager=MockMongoWithOpenSearchManager).items() if k is not None})

    def insert_content(self, content: Optional[Mapping[CollectionKey, Mapping[str, bytes]]]):
        super().insert_content(content)
        assert self.__mongo_opensearch is not None
        if content:
            self.__mongo_opensearch.add_content({k: v for k, v in query_content(content, db_manager=MockMongoWithOpenSearchManager).items() if k is not None})

    def get_database(self) -> MockMongoWithOpenSearch:
        if self.__mongo_opensearch is None:
            raise ValueError('start_database not called')
        return self.__mongo_opensearch


class MockS3WithOpenSearch(MockDatabase, InMemoryDatabase, S3WithOpenSearch):
    """
    Mock implementation of the S3WithOpenSearch class.
    """
    def __init__(self, config: Optional[configparser.ConfigParser] = None,
                 s3_opensearch: Optional['MockS3WithOpenSearch'] = None,
                 **kwargs) -> None:
        super().__init__(config, **kwargs)
        if s3_opensearch is not None:
            self.add_desktop_objects(s3_opensearch.get_all_desktop_objects())
            self.add_content(s3_opensearch.get_all_content())

    async def search(self,query: dict, search_item_type: type[ItemTypeVar],
                     index: str | None = None,
                     max_results: int = 1000,
                     size: int = 100) -> list[ItemTypeVar]:
        return [search_item_type()]

    async def ping(self):
        pass


class MockS3WithOpenSearchManager(OpenSearchManager):
    """
    Mock manager for S3WithOpenSearch that stores collections in memory.
    """
    def __init__(self) -> None:
        super().__init__()
        self.__s3_opensearch: Optional[MockS3WithOpenSearch] = None

    @classmethod
    def get_context(cls) -> list[AbstractContextManager]:
        result = super().get_context()
        result.append(freeze_time())
        return result

    def start_database(self, context_manager: ExitStack) -> None:
        self.__s3_opensearch = MockS3WithOpenSearch()
        super().start_database(context_manager)

    def insert_desktop_objects(self, desktop_objects: Optional[Mapping[CollectionKey, Sequence[DesktopObjectDict]]]):
        super().insert_desktop_objects(desktop_objects)
        assert self.__s3_opensearch is not None
        if desktop_objects:
            self.__s3_opensearch.add_desktop_objects({k: v for k, v in query_fixtures(desktop_objects, db_manager=MockS3WithOpenSearchManager).items() if k is not None})

    def insert_content(self, content: Optional[Mapping[CollectionKey, Mapping[str, bytes]]]):
        super().insert_content(content)
        assert self.__s3_opensearch is not None
        if content:
            self.__s3_opensearch.add_content({k: v for k, v in query_content(content, db_manager=MockS3WithOpenSearchManager).items() if k is not None})

    def get_database(self) -> MockS3WithOpenSearch:
        if self.__s3_opensearch is None:
            raise ValueError('start_database not called')
        return self.__s3_opensearch



"""
Utility classes and functions for working with aiohttp.
"""
import asyncio
from contextlib import AbstractContextManager
from multiprocessing.connection import Connection
from io import BytesIO
from aiohttp.streams import StreamReader
from aiohttp.web import Request, HTTPError
from aiohttp.client import ClientSession
from typing import Protocol, AsyncIterator, BinaryIO
from heaobject.root import json_dumps
from heaobject.user import NONE_USER
from .oidcclaimhdrs import SUB
from .appproperty import HEA_COMPONENT
from contextlib import asynccontextmanager
import logging
import os
import time
from typing import ParamSpec, Callable

from yarl import URL


CHUNK_SIZE = 16 * 1024


@asynccontextmanager
async def client_session(*args, **kwargs) -> AsyncIterator[ClientSession]:
    """
    Gets a new aiohttp ClientSession. The session's json_serialize parameter is set to heaobject.root.json_dumps,
    which unlike Python's native json serialization supports serializing timestamps. It supports any arguments that
    ClientSession's constructor supports.
    """
    session = ClientSession(json_serialize=json_dumps, *args, **kwargs)
    try:
        yield session
    finally:
        await session.close()


class ConnectionFileLikeObjectWrapper(AbstractContextManager):
    """
    Wraps a multiprocessing.connection.Connection object and provides file-like object methods.

    This class is a context manager, so it can be used in with statements.
    """

    def __init__(self, conn: Connection):
        """
        Creates a new ConnectionFileLikeObjectWrapper object, passing in the connection to wrap.

        :param conn: a multiprocessing.connection.Connection object (required).
        """
        if conn is None:
            raise ValueError('conn cannot be None')
        self.__conn = conn
        self.__buffer = BytesIO()

    def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        if len(b := self.__buffer.read(n)) > 0:
            return b
        try:
            result = self.__conn.recv_bytes()
            if -1 < n < len(result):
                pos = self.__buffer.tell()
                self.__buffer.write(result[n:])
                self.__buffer.seek(pos)
                return result[:n]
            else:
                return result
        except EOFError:
            return b''

    def write(self, b):
        """
        Sends some bytes to the connection.

        :param b: some bytes (required).
        """
        self.__conn.send_bytes(b)

    def fileno(self) -> int:
        """
        Returns the integer file descriptor that is used by the connection.

        :return: the integer file descriptor.
        """
        return self.__conn.fileno()

    def close(self) -> None:
        """
        Closes the connection and any other resources associated with this object.
        """
        try:
            self.__buffer.close()
            self.__conn.close()
        finally:
            if not self.__conn.closed:
                try:
                    self.__conn.close()
                except OSError:
                    pass

    def __exit__(self, *exc_details):
        self.close()


class SupportsAsyncRead(Protocol):
    """
    Protocol with an async read() method and a close() method.
    """

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        pass

    def close(self):
        """
        Closes any resources associated with this object.
        """
        pass


class AsyncReader:
    """
    Wraps a bytes object in a simple reader with an asynchronous read method and a close method.
    """

    def __init__(self, b: bytes):
        """
        Creates a new AsyncReader, passing in a bytes object.

        :param b: bytes (required).
        """
        self.__b = BytesIO(b)

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return self.__b.read(n)

    def close(self):
        """
        Closes any resources associated with this object.
        """
        self.__b.close()


class StreamReaderWrapper:
    """
    Wraps an aiohttp StreamReader in an asyncio StreamReader-like object with a read() method and a no-op close()
    method.
    """

    def __init__(self, reader: StreamReader):
        if reader is None:
            raise ValueError('reader cannot be None')
        self.__reader = reader

    async def read(self, n=-1):
        """
        Reads up to n bytes. If n is not provided, or set to -1, reads until EOF and returns all read bytes.

        If the EOF was received and the internal buffer is empty, returns an empty bytes object.

        :param n: how many bytes to read, -1 for the whole stream.
        :return: the data.
        """
        return await self.__reader.read(n)

    def close(self):
        pass


class RequestFileLikeWrapper:
    """
    Wraps an aiohttp request's content in a file-like object with read() and close() functions. Before doing any
    reading, call the initialize() method. The read() method must be called in a separate thread.
    """

    def __init__(self, request: Request, loop: asyncio.AbstractEventLoop | None = None) -> None:
        """
        Creates the file-like object wrapper.

        :param request: the aiohttp request (required).
        :param loop: the current event loop. If None, it will use asyncio.get_running_loop().
        """
        if loop is not None:
            self.loop = loop
        else:
            self.loop = asyncio.get_running_loop()
        self.request = request
        self.__pump_task: asyncio.Task | None = None
        self.__reader: BinaryIO | None = None
        self.__writer: BinaryIO | None = None

    def initialize(self):
        """
        Creates resources needed for reading.
        """
        read_fd, write_fd = os.pipe()
        self.__reader = os.fdopen(read_fd, 'rb')
        self.__writer = os.fdopen(write_fd, 'wb')
        self.__pump_task = self.loop.create_task(self.__pump_bytes_into_fd())

    def read(self, n=-1, /) -> bytes:
        """
        Reads some bytes.

        :param n: the number of bytes to read (or -1 for everything).
        :return: the bytes that were read.
        """
        logger = logging.getLogger(__name__)
        logger.debug('Reading %s bytes', n)
        assert self.__reader is not None, 'Must call initialize() before calling read()'
        output = self.__reader.read(n)
        logger.debug('Read %s bytes', len(output))
        return output

    def close(self):
        """
        Cleans up all resources in this file-like object.
        """
        logger = logging.getLogger(__name__)
        logger.debug('Closing')
        try:
            if self.__writer is not None:
                self.__writer.close()
                self.__writer = None
            if self.__reader is not None:
                self.__reader.close()
                self.__reader = None
            while self.__pump_task is not None and not self.__pump_task.done():
                time.sleep(0.1)
            self.__pump_task = None
        except Exception as e:
            logger.exception('Failed to close pipe')
            if self.__reader is not None:
                try:
                    self.__reader.close()
                    self.__reader = None
                except:
                    pass
            try:
                while self.__pump_task is not None and not self.__pump_task.done():
                    time.sleep(0.1)
                self.__pump_task = None
            except:
                pass
            raise e

    async def __pump_bytes_into_fd(self):
        logger = logging.getLogger(__name__)
        writer_closed = False
        content_type = self.request.headers.get('Content-Type', '')
        try:
            if content_type.startswith('multipart/form-data'):
                logger.debug(f'Is multipart')
                mp_reader = await self.request.multipart()
                while True:
                    part = await mp_reader.next()
                    if part is None:
                        break
                    logger.debug("the part's headers %s" % part.headers)
                    while True:
                        chunk = await part.read_chunk(CHUNK_SIZE)
                        if not chunk or chunk == b'':
                            break
                        logger.debug('Read %d bytes from upload', len(chunk))
                        bytes_written = await self.loop.run_in_executor(None, self.__writer.write, chunk)
                        logger.debug('Wrote %d bytes to pipe', bytes_written)
            else:
                while not self.request.content.at_eof():
                    chunk = await self.request.content.read(CHUNK_SIZE)
                    logger.debug('Read %d bytes from upload', len(chunk))
                    bytes_written = await self.loop.run_in_executor(None, self.__writer.write, chunk)
                    logger.debug('Wrote %d bytes to pipe', bytes_written)
            self.__writer.close()
            writer_closed = True
            logger.debug('Done reading file')
        except Exception as e:
            logger.exception('Failed to read file')
            if not writer_closed:
                try:
                    self.__writer.close()
                except:
                    pass
                writer_closed = False
                raise e

def absolute_url_minus_base(request: Request, url: URL | str | None = None) -> str:
    """
    Removes the current service's base URL from the beginning of the provided URL. If no URL is provided, the request
    URL is used. If the provided URL is not prefixed with the service's base URL, the URL is returned unaltered.

    :param request: the HTTP request (required).
    :param url: optional URL.
    :return: the resulting URL.
    """
    if url is None:
        url_ = str(request.url)
    else:
        url_ = str(url)
    base_url = request.app[HEA_COMPONENT]
    if not url_.startswith(base_url):
        return url_
    else:
        return url_.removeprefix(base_url).lstrip('/')


P = ParamSpec('P')

def http_error_message(http_error: HTTPError, object_display_name: Callable[P, str], *args: P.args, **kwargs: P.kwargs) -> HTTPError:
    """
    If the HTTPError object has an empty body, it will try filling in the body with a message appropriate for the given
    status code for operations on desktop objects.

    :param http_error: the HTTPError (required). This object's body may be modified by this call if it is None or an
    empty byte array.
    :param object_display_name: a callable that generates a description of a desktop object for inclusion in an error
    message.
    :param args: positional arguments to pass to object_display_name.
    :param kwargs: keyword arguments to pass to object_display_name. The encoding keyword argument is reserved and will
    be used to set the encoding of the error message when encoded to bytes (the default is utf-8).
    :return: the updated HTTPError (the same object as the http_error argument).
    """
    encoding = kwargs.get('encoding')
    if encoding:
        encoding_ = str(encoding)
    else:
        encoding_ = 'utf-8'
    if not http_error.body:
        match http_error.status:
            case 403:
                http_error.body = f'Access to {object_display_name(*args, **kwargs)} is denied because you do not have sufficient permission for it'.encode(encoding=encoding_)
            case 404:
                http_error.body = f'{object_display_name(*args, **kwargs)} not found'.encode(encoding=encoding_)
    return http_error


def extract_sub(request: Request) -> str:
    """
    Extracts the current user (subject) from the request's OIDC_CLAIM_sub header.

    :param request: the HTTP request (required).
    :return: the user id, or system|none if the header has no value or is absent.
    """
    return request.headers.get(SUB, NONE_USER)



import warnings

from heaobject.folder import AWSS3Folder, AWSS3ItemInFolder
from heaserver.service.expression import get_eval_for, EvaluatorException
from unittest import TestCase
from typing import Any
from typing import NamedTuple


class BoolExpressionTuple(NamedTuple):
    expr: str
    value: bool


class ExceptionExpressionTuple(NamedTuple):
    expr: str
    value: type[Exception]

# Object passed in is an AWSS3Folder.
TEST_EXPRESSIONS = [
        BoolExpressionTuple(expr='isheaobject(heaobject.folder.AWSS3Folder)', value=True),
        BoolExpressionTuple('display_name == "foobar"', True),
        BoolExpressionTuple('id is not None', True),
        BoolExpressionTuple('id is None or display_name is None', False),
        BoolExpressionTuple('"this" in description', True)
    ]

# Object passed in is an AWSS3Folder; an AWSS3ItemInFolder is passed in as an extra name obj2.
TEST_EXPRESSIONS_WITH_EXTRA = [
        BoolExpressionTuple(expr='isheaobject(heaobject.folder.AWSS3ItemInFolder)', value=False),
        BoolExpressionTuple('"this" in obj2.description if obj2.description else False', False),
        BoolExpressionTuple('display_name == obj2.display_name', True),
        BoolExpressionTuple('description == obj2.description', False),
        BoolExpressionTuple('display_name is None and description is None', False)
    ]

TEST_EXPRESSIONS_THROWING_EXCEPTIONS = [
    ExceptionExpressionTuple('notaname is None', EvaluatorException),
    ExceptionExpressionTuple('4 of None', SyntaxError),
    ExceptionExpressionTuple('notaname of None', SyntaxError),
    ExceptionExpressionTuple('isinstance(obj2, heaobject.folder.AWSS3Folder)', EvaluatorException),
    ExceptionExpressionTuple('def foo: pass', SyntaxError),
    ExceptionExpressionTuple('def foo: print(notaname)', SyntaxError),
    ExceptionExpressionTuple('def foo(): pass', EvaluatorException),
    ExceptionExpressionTuple('class Foo: pass', EvaluatorException),
    ExceptionExpressionTuple('class Foo: a = notaname', EvaluatorException),
    ExceptionExpressionTuple('foo = "bar"', EvaluatorException),
    ExceptionExpressionTuple('if 4 == 3: print(5)', EvaluatorException),
    ExceptionExpressionTuple('while True: print("hello world")', EvaluatorException)
]

class TestEvaluator(TestCase):

    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        self.eval_ = get_eval_for(obj).eval

    def test_empty_string(self):
        with self.assertRaises(EvaluatorException):
            self.eval_('')

    def test_None_expression(self):
        with self.assertRaises(EvaluatorException):
            self.eval_(None)

def _my_test_generator(expr: str, expected: Any):
    def test_(self):
        self.assertEqual(expected, self.eval_(expr))
    return test_


for test_case in TEST_EXPRESSIONS:
    test_case_name = test_case.expr if test_case.expr else 'empty_expression'
    setattr(TestEvaluator, f'test_{test_case_name}', _my_test_generator(test_case.expr, test_case.value))


class TestEvaluatorDict(TestCase):

    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        self.eval_ = get_eval_for(obj.to_dict()).eval


for test_case in TEST_EXPRESSIONS:
    test_case_name = test_case.expr if test_case.expr else 'empty_expression'
    setattr(TestEvaluatorDict, f'test_{test_case_name}', _my_test_generator(test_case.expr, test_case.value))


class TestEvaluatorWithExtra(TestCase):

    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        obj2 = AWSS3ItemInFolder()
        obj2.display_name = 'foobar'

        self.eval_ = get_eval_for(obj, extra_names={'obj2': obj2}).eval


def _my_test_generator_with_extra(expr: str, expected: Any):
    def test_(self: TestEvaluatorWithExtra):
        self.assertEqual(expected, self.eval_(expr))
    return test_


for test_case in TEST_EXPRESSIONS_WITH_EXTRA:
    setattr(TestEvaluatorWithExtra, f'test_{test_case.expr}', _my_test_generator_with_extra(test_case.expr, test_case.value))


class TestEvaluatorWithExtraDict(TestCase):

    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        obj2 = AWSS3ItemInFolder()
        obj2.display_name = 'foobar'

        self.eval_ = get_eval_for(obj.to_dict(), extra_names={'obj2': obj2}).eval


for test_case in TEST_EXPRESSIONS_WITH_EXTRA:
    setattr(TestEvaluatorWithExtraDict, f'test_{test_case.expr}', _my_test_generator_with_extra(test_case.expr, test_case.value))



class TestEvaluatorThrowingExceptions(TestCase):
    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        obj2 = AWSS3ItemInFolder()
        obj.display_name = 'foobar'

        self.eval_ = get_eval_for(obj, extra_names={'obj2': obj2}).eval


def _my_test_generator_throwing_exceptions(expr: str, expected: type[Exception]):
    def test_(self):
        with self.assertRaises(expected):
            self.eval_(expr)
    return test_


for test_case in TEST_EXPRESSIONS_THROWING_EXCEPTIONS:
    setattr(TestEvaluatorThrowingExceptions, f'test_{test_case.expr}', _my_test_generator_throwing_exceptions(test_case.expr, test_case.value))


class TestEvaluatorThrowingExceptionsDict(TestCase):
    def setUp(self) -> None:
        obj = AWSS3Folder()
        obj.display_name = 'foobar'
        obj.description = 'this is a description'

        obj2 = AWSS3ItemInFolder()
        obj.display_name = 'foobar'

        self.eval_ = get_eval_for(obj.to_dict(), extra_names={'obj2': obj2}).eval


for test_case in TEST_EXPRESSIONS_THROWING_EXCEPTIONS:
    setattr(TestEvaluatorThrowingExceptionsDict, f'test_{test_case.expr}', _my_test_generator_throwing_exceptions(test_case.expr, test_case.value))



"""
Documentation for setup.py files is at https://setuptools.readthedocs.io/en/latest/setuptools.html
"""

from setuptools import setup, find_namespace_packages

# Import the README.md file contents
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='heaserver',
      version='1.27.2',
      description='The server side of HEA.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://risr.hci.utah.edu',
      author='Research Informatics Shared Resource, Huntsman Cancer Institute, Salt Lake City, UT',
      author_email='Andrew.Post@hci.utah.edu',
      python_requires='>=3.10',
      package_dir={'': 'src'},
      packages=find_namespace_packages(where='src'),
      package_data={'heaserver.service': ['py.typed', 'jsonschemafiles/*']},
      license='Apache License 2.0',
      install_requires=[
          'heaobject~=1.20.0',
          'aiohttp~=3.10.10',  # aiohttp-swagger3 doesn't support aiohttp 3.11 yet.
          'aiodns; platform_system!="Windows"',  # Optional aiohttp speedup; requires the older Windows event loop, so disable this dependency on Windows for now.
          'Brotli',  # Optional aiohttp speedup.
          'aiohttp-remotes~=1.3.0',
          'motor~=3.6.0',
          'motor-types~=1.0.0b4',
          'accept-types~=0.4.1',
          'mongoquery~=1.4.2',
          'jsonschema~=4.23.0',
          'jsonmerge~=1.9.2',
          'requests~=2.31.0',  # Upgrade to 2.32.* when docker >= 7.1.0.
          'types-requests>=2.32.0.20241016',  # Should be set at same version as requests.
          'boto3[crt]~=1.37.25',
          'botocore~=1.37.25',
          'boto3-stubs[essential,sts,account,organizations,iam]~=1.37.25',
          'botocore-stubs[essential]~=1.37.25',
          'freezegun~=1.5.1',
          'regex~=2024.11.6',
          'aio-pika==9.5.1',
          'simpleeval~=1.0.3',
          'opensearch-py[async]~=2.6.0',
          'cachetools~=5.5.0',
          'types-cachetools~=5.5.0.20240820',
          'types-pywin32; platform_system=="Windows"',
          'uvloop~=0.21.0; platform_system!="Windows"',
          'aiodiskqueue~=0.1.2',
          'aiosqlite~=0.21.0'
      ],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Software Development',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Topic :: Scientific/Engineering :: Medical Science Apps.'
      ]
      )



"""
This module defines functions that create HTTP responses.

All GET responses except for OPTIONS include Cache-Control, Pragma, and Expires headers to request no caching.
"""
from aiohttp import web, hdrs
from aiohttp.client_exceptions import ClientResponseError
from aiohttp.helpers import ETag
from aiohttp.typedefs import LooseHeaders
from hashlib import md5
from heaobject.user import NONE_USER

from .representor.factory import formats_from_request
from . import requestproperty, appproperty
from .appproperty import HEA_BACKGROUND_TASKS
from .oidcclaimhdrs import SUB
from .representor.representor import Link
from yarl import URL
import logging
from typing import Union, Optional, Any
from collections.abc import Iterable, Mapping, Sequence
from .aiohttp import SupportsAsyncRead
from .wstl import RuntimeWeSTLDocumentBuilder
from heaobject.root import DesktopObjectDict, Permission
from multidict import istr, CIMultiDict
from .caching_strategy import CachingStrategy

TEXT_PLAIN_HTTP_HEADERS: Mapping[Union[str, istr], str] = CIMultiDict({hdrs.CONTENT_TYPE: 'text/plain; charset=UTF-8'})
NO_CACHE_HEADERS = CIMultiDict({hdrs.CACHE_CONTROL: 'no-cache, no-store, must-revalidate', hdrs.PRAGMA: 'no-cache', hdrs.EXPIRES: '0'})
DEFAULT_CONTENT_TYPE = 'application/json'


def status_generic(status: int, body: Optional[Union[bytes, str]] = None,
                   headers: LooseHeaders | None = None) -> web.Response:
    """
    Returns a newly created HTTP response object. This object is not an exception. To return an error response that
    can be raised as an exception, use status_generic_error().

    :param status: an HTTP status code (required).
    :param body: the response body (optional).
    :param headers: optional headers.
    :return: an HTTP response.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.Response(status=status, body=body_, headers=headers)


def status_generic_error(status: int, body: Optional[Union[bytes, str]] = None,
                         headers: LooseHeaders | None = None) -> web.HTTPError:
    """
    Returns a newly created HTTP error response object. This object is an exception.

    :param status: an HTTP status code (required).
    :param body: the response body (optional).
    :param headers: optional headers.
    :return: an HTTPError that can be used a HTTP response or an exception.
    """
    if status < 400:
        raise ValueError(f'status must be an error code >=400 but was {status}')
    body_ = body.encode() if isinstance(body, str) else body
    class HTTPGenericError(web.HTTPError):
        status_code = status
    return HTTPGenericError(body=body_, headers=headers)


def status_not_found(body: Optional[Union[bytes, str]] = None,
                     headers: LooseHeaders | None = None) -> web.HTTPNotFound:
    """
    Returns a newly created HTTP response object with status code 404 and an optional body.

    :param body: the body of the message, typically an explanation of what was not found.  If a string, this function
    will encode it to bytes using UTF-8 encoding.
    :param headers: any headers that you want to add to the response. In addition, the response will have a
    Content-Type header that should not be overridden.
    :return: aiohttp.web.Response with a 404 status code.
    """
    if headers and 'Content-Type' in headers:
        raise ValueError('The Content-Type header cannot be overridden')
    headers_: CIMultiDict[str] = CIMultiDict(headers) if headers is not None else CIMultiDict()
    if isinstance(body, str):
        headers_.update(TEXT_PLAIN_HTTP_HEADERS)
    return web.HTTPNotFound(body=body.encode() if isinstance(body, str) else body, headers=headers_)


def status_multiple_choices(default_url: Union[URL, str], body: Optional[Union[bytes, str]] = None,
                            content_type: str = DEFAULT_CONTENT_TYPE) -> web.HTTPMultipleChoices:
    """
    Returns a newly created HTTP response object with status code 300. This is for implementing client-side content
    negotiation, described at https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation.

    :param default_url: the URL of the default choice. Required.
    :param body: content with link choices.  If a string, this function will encode it to bytes using UTF-8 encoding.
    :param content_type: optional content_type (defaults to DEFAULT_CONTENT_TYPE). Cannot be None.
    :return: aiohttp.web.Response object with a 300 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.HTTPMultipleChoices(str(default_url),
                                   body=body_,
                                   headers={hdrs.CONTENT_TYPE: content_type,
                                            hdrs.LOCATION: str(default_url if default_url else '#')})


def status_moved(location: str | URL) -> web.HTTPMovedPermanently:
    """
    Returns a newly created HTTP response object with status code 301.

    :param location: the URL for the client to redirect to.
    :return: aiohttp.web.Response object with a 301 status code.
    """
    return web.HTTPMovedPermanently(location=str(location))


def status_bad_request(body: Optional[Union[bytes, str]] = None) -> web.HTTPBadRequest:
    """
    Returns a newly created HTTP response object with status code 400 and an optional body.

    :param body: the body of the message, typically an explanation of why the request is bad.  If a string, this
    function will encode it to bytes using UTF-8 encoding.
    :return: aiohttp.web.Response with a 400 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    headers = TEXT_PLAIN_HTTP_HEADERS if isinstance(body, str) else None
    return web.HTTPBadRequest(body=body_, headers=headers)


def status_created(base_url: Union[URL, str], resource_base: str, inserted_id: str) -> web.HTTPCreated:
    """
    Returns a newly created HTTP response object with status code 201 and the Location header set.

    :param base_url: the service's base URL (required).
    :param resource_base: the common base path fragment for all resources of this type (required).
    :param inserted_id: the id of the newly created object (required).

    :return: aiohttp.web.Response with a 201 status code and the Location header set to the URL of the created object.
    """
    if inserted_id is None:
        raise ValueError('inserted_id cannot be None')
    return web.HTTPCreated(headers={hdrs.LOCATION: str(URL(base_url) / resource_base / str(inserted_id))})


def status_ok(body: Optional[Union[bytes, str]] = None, content_type: str = DEFAULT_CONTENT_TYPE,
              headers: LooseHeaders | None = None) -> web.HTTPOk:
    """
    Returns a newly created HTTP response object with status code 201, the provided Content-Type header, and the
    provided body.

    :param body: the body of the response. If a string, this function will encode it to bytes using UTF-8 encoding.
    :param content_type: the content type of the response (default is application/json).
    :param headers: any headers that you want to add to the response, in addition to the content type. If the content
    type is also in this dictionary, then its value will override that of the content_type argument. If the body is
    a string and no content type header is specified in either the content_type or headers argument, a default value
    for the content type will be used.
    :return: aiohttp.web.Response object with a 200 status code.
    """
    if isinstance(body, str):
        body_: bytes | None = body.encode()
    else:
        body_ = body
    headers_ = CIMultiDict(headers or {})
    if content_type is not None:
        headers_.update({hdrs.CONTENT_TYPE: content_type})
    if isinstance(body, str) and hdrs.CONTENT_TYPE not in headers_:
        headers_.update(TEXT_PLAIN_HTTP_HEADERS)
    return web.HTTPOk(headers=headers_, body=body_)


def status_no_content() -> web.HTTPNoContent:
    """
    Returns a newly created HTTP response object with status code 204.

    :return: aiohttp.web.Response object with a 204 status code.
    """
    return web.HTTPNoContent()


def status_not_acceptable(body: Optional[Union[bytes, str]] = None) -> web.HTTPNotAcceptable:
    """
    Returns a newly created HTTP response object with status code 406.

    :param body: the body of the message, typically an explanation of why the request is bad.  If a string, this
    function will encode it to bytes using UTF-8 encoding.
    :return: aiohttp.web.Response object with a 406 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.HTTPNotAcceptable(body=body_)


def status_internal_error(body: Optional[Union[bytes, str]] = None) -> web.HTTPInternalServerError:
    """
    Returns a newly created HTTP response object with status code 500.

    :param body: the body of the message, typically an explanation of why the request is bad.  If a string, this
    function will encode it to bytes using UTF-8 encoding.
    :return: aiohttp.web.Response object with a 500 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.HTTPInternalServerError(body=body_)


def status_from_exception(e: ClientResponseError) -> web.HTTPError:
    """
    Generates a response from an aiohttp ClientResponseError, using the status code and message from the ClientResponse
    object.

    :param e: a ClientResponseError object.
    :return: an web.HTTPError object with the ClientResponseError's status code and message.
    """
    return status_generic_error(status=e.status, body=e.message)


def status_forbidden(body: Optional[Union[bytes, str]] = None) -> web.HTTPForbidden:
    """
    Returns a newly created HTTP response object with status code 403.

    :param body: the body of the message, typically an explanation of why the request is bad.  If a string, this
    function will encode it to bytes using UTF-8 encoding.
    :return: aiohttp.web.Response object with a 403 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.HTTPForbidden(body=body_)


def status_see_other(location: URL | str) -> web.Response:
    return web.HTTPSeeOther(location)


def status_accepted() -> web.Response:
    return web.HTTPAccepted()


def status_conflict(body: Optional[Union[bytes, str]] = None) -> web.HTTPConflict:
    """
    Returns a newly created HTTP response object with status code 409.

    :param body: the body of the message, typically an explanation of why the request is in conflict.  If a string,
    this function will encode it to bytes using UTF-8 encoding.
    :return: aiohttp.web.Response object with a 409 status code.
    """
    body_ = body.encode() if isinstance(body, str) else body
    return web.HTTPConflict(body=body_)


async def get(request: web.Request, data: Optional[DesktopObjectDict], permissions: Sequence[Permission] | None = None,
              attribute_permissions: Mapping[str, Sequence[Permission]] | None = None,
              include_data = True) -> web.Response:
    """
    Create and return a HTTP response object in response to a GET request for one or more HEA desktop object resources.
    The request headers are used to select a format for including the data and associated metadata such as links and
    permissions in the response body. See the wstl module and representor package for details.

    :param request: the HTTP request (required).
    :param data: a HEA desktop object dict. If None or an empty dictionary, the response will have a 404 Not Found
    status code.
    :param permissions: the object-level permissions for the desktop object. None or empty means that the permissions
    for the desktop object are unknown.
    :param attribute_permissions: the attribute-level permissions for the desktop object. None or empty means that the
    permissions for the desktop object are unknown.
    :param include_data: whether to include the data in the response. If False, and the data parameter is not None and
    a non-empty dictionary, a 200 status code is returned and the data is omitted, but links are included in the
    response body. Ensure that the request parameters and data parameter contain any information needed to generate the
    links. If include_data is False, the permissions and attribute_permissions parameters are ignored and should be
    omitted.
    :return: aiohttp.web.Response object, with status code 200, containing a body with the HEA desktop object, or
    status code 404 if the data argument is None.
    """
    if data:
        return await _handle_get_result(request, data, [permissions] if permissions is not None else None,
                                        [attribute_permissions] if attribute_permissions is not None else None,
                                        include_data=include_data)
    else:
        return web.HTTPNotFound()


async def get_multiple_choices(request: web.Request,
                               result: DesktopObjectDict | None = None,
                               caching_strategy: CachingStrategy | None = None) -> web.Response:
    """
    Create and return a HTTP response to a GET request with available links for opening the requested desktop
    object. Unlike with typical GET requests, this function generates a successful response with status code 300 to
    indicate that the endpoint is for the purpose of client-side content negotiation. More information about content
    negotiation is available from https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation.

    The format of the links in the response body depends on the the result of server-side content negotiation using the
    request's Accept header and the REST endpoint's supported content types. The default content type is
    Collection+JSON.

    Multiple choice REST endpoints defining actions that have parameters not found in the endpoint's route require
    passing a desktop object into this function's result parameter. Failing to provide a desktop object for an endpoint
    requiring one has undefined behavior. For endpoints that do not require a desktop object, omitting one for an
    otherwise correct request will result in a successful response with a 300 status code and the options for opening
    the requested object, depending on the response body format negotiated between the client and server. To return an
    error status code, use the status_not_found() call or similar instead of this function.

    :param request: the HTTP request (required). If an Accepts header is provided, MIME types that do not support
    links will be ignored.
    :param result: a HEA desktop object dict (optional). As described above, some REST endpoints require this
    parameter to have a non-None value, otherwise the behavior is undefined. For REST endpoints that do not require
    a non-None value, a successful response should have a 300 status code.
    :param caching_strategy: the caching strategy to use. If None, then response headers are sent to request no caching.
    :return: aiohttp.web.Response object with status code 300, and a body containing the links representing possible
    choices for opening the HEA desktop object; status code 404 if no HEA desktop object dict was provided; or status
    code 406 if content negotiation failed to determine an acceptable content type.
    """
    _logger = logging.getLogger(__name__)
    wstl_builder: RuntimeWeSTLDocumentBuilder = request[requestproperty.HEA_WSTL_BUILDER]
    wstl_builder.href = str(request.url)

    default_url: Optional[Union[URL, str]] = None

    def link_callback(action_index: int, link: Link):
        nonlocal default_url
        if default_url is None or 'default' in link.rel:
            default_url = link.href
    try:
        body, mime_type = await wstl_builder.represent_from_request(request,
                                                                    [result] if result else None,
                                                                    link_callback=link_callback,
                                                                    include_data=False)
        _logger.debug('Response body is %s', body)
        etag = _compute_etag(body)
        if request.if_none_match and etag in request.if_none_match:
            return web.HTTPNotModified()
        response = status_multiple_choices(default_url=default_url if default_url else '#', body=body,
                                           content_type=mime_type)
        if caching_strategy:
            response.etag = etag
        else:
            response.headers.update(NO_CACHE_HEADERS)
        return response
    except ValueError as e:
        return status_not_acceptable(str(e))


async def get_from_wstl(request: web.Request, run_time_doc: dict[str, Any]) -> web.Response:
    """
    Handle a get request that returns a run-time WeSTL document. Any actions in the document are added to the
    request's run-time WeSTL documents, and the href of the action is prepended by the service's base URL. The actions
    in the provided run-time document are expected to have a relative href.

    :param request: the HTTP request (required).
    :param run_time_doc: a run-time WeSTL document containing data.
    :return: aiohttp.web.Response object with a body containing the object in a JSON array of objects.
    """
    return await _handle_get_result_from_wstl(request, [run_time_doc])


async def get_all_from_wstl(request: web.Request, run_time_docs: Iterable[dict[str, Any]]) -> web.Response:
    """
    Handle a get all request that returns one or more run-time WeSTL documents. Any actions in the documents are added
    to the request's run-time WeSTL documents, and the href of the action is prepended by the service's base URL. The
    actions in the provided run-time document are expected to have a relative href.

    :param request: the HTTP request (required).
    :param run_time_docs: a list of run-time WeSTL documents containing data.
    :return: aiohttp.web.Response object with a body containing the object in a JSON array of objects.
    """
    return await _handle_get_result_from_wstl(request, run_time_docs)


async def get_all(request: web.Request, data: Sequence[DesktopObjectDict],
                  permissions: Sequence[Sequence[Permission]] | None = None,
                  attribute_permissions: Sequence[Mapping[str, Sequence[Permission]]] | None = None) -> web.Response:
    """
    Create and return a Response object in response to a GET request for all HEA desktop object resources in a
    collection.

    :param request: the HTTP request (required).
    :param data: a list of HEA desktop object dicts.
    :return: aiohttp.web.Response object with a body containing the object in a JSON array of objects.
    """
    return await _handle_get_result(request, data, permissions, attribute_permissions)


async def get_options(request: web.Request, methods: Iterable[str]) -> web.Response:
    """
    Create and return a Response object in response to an OPTIONS request.

    :param request: the HTTP request (required).
    :param methods: the allowed HTTP methods.
    :return: an aiohttp.web.Response object with a 200 status code and an Allow header.
    """
    methods_ = ', '.join(methods)
    etag = _compute_etag(methods_)
    if request.if_none_match and etag in request.if_none_match:
        return web.HTTPNotModified()
    resp = web.HTTPOk()
    resp.headers[hdrs.ALLOW] = methods_
    return resp


async def get_streaming(request: web.Request, out: SupportsAsyncRead, content_type: str = DEFAULT_CONTENT_TYPE,
                        caching_strategy: CachingStrategy | None = None) -> web.StreamResponse:
    """
    Create and return a StreamResponse object in response to a GET request for the content associated with a HEA desktop
    object.

    :param request: the HTTP request (required).
    :param out: a file-like object with an asynchronous read() method (required). The stream is closed after all data
    is read from it.
    :param content_type: optional content type.
    :param caching_strategy: the caching strategy to employ. If unspecified, response headers are sent to request no
    caching.
    :return: aiohttp.web.StreamResponse object with status code 200.
    """
    logger = logging.getLogger(__name__)
    logger.debug('Getting content with content type %s', content_type)
    if content_type is not None:
        response_ = web.StreamResponse(status=200, reason='OK', headers={hdrs.CONTENT_TYPE: content_type})
    else:
        response_ = web.StreamResponse(status=200, reason='OK')
    await response_.prepare(request)
    try:
        while chunk := await out.read(1024):
            await response_.write(chunk)
    finally:
        out.close()

    await response_.write_eof()
    if not caching_strategy:
        response_.headers.update(NO_CACHE_HEADERS)
    return response_


async def post(request: web.Request, result: Optional[str], resource_base: str) -> web.Response:
    """
    Create and return a Response object in response to a POST request to create a new HEA desktop object resource.

    :param request: the HTTP request (required).
    :param result: the id of the POST'ed HEA object, or None if the POST failed due to a bad request.
    :param resource_base: the common base path for all resources of this type (required).
    :return: aiohttp.web.Response with Created status code and the URL of the created object, or a Response with a Bad
    Request status code if the result is None.
    """
    logger = logging.getLogger(__name__)
    if result is not None:
        return await _handle_post_result(request, resource_base, result)
    else:
        logger.debug('No result for request %s', request.url)
        return web.HTTPBadRequest()


async def put(result: bool) -> web.Response:
    """
    Handle the result from a put request.

    :param result: whether any objects were updated.
    :return: aiohttp.web.Response object with status code 203 (No Content) or 404 (Not Found).
    """
    if result:
        return web.HTTPNoContent()
    else:
        return web.HTTPNotFound()


async def delete(result: bool) -> web.Response:
    """
    Handle the result from a delete request.

    :param result: whether any objects were deleted.
    :return: aiohttp.web.Response object with status code 203 (No Content) or 404 (Not Found).
    """
    if result:
        return web.HTTPNoContent()
    else:
        return web.HTTPNotFound()


async def _handle_get_result(request: web.Request, data: Union[DesktopObjectDict, Sequence[DesktopObjectDict]],
                             permissions: Sequence[Sequence[Permission]] | None = None,
                             attribute_permissions: Sequence[Mapping[str, Sequence[Permission]]] | None = None,
                             include_data = True) -> web.Response:
    """
    Handle the result from a get request. Returns a Response object, the body of which will always contain a list of
    JSON objects.

    :param request: the HTTP request object. Cannot be None.
    :param data: the retrieved HEA desktop objects as a dict or a list of dicts, with each dict representing a
    desktop object with its attributes as name-value pairs.
    :param include_data: whether to include the data in the response. If False, and the data parameter is not None and
    a non-empty dictionary, a 200 status code is returned and the data is omitted, but links are included in the
    response body. Ensure that the request parameters and data parameter contain any information needed to generate the
    links. If include_data is False, the permissions and attribute_permissions parameters are ignored and should be
    omitted.
    :return: aiohttp.web.Response object, either a 200 status code and the requested JSON object in the body, or status
    code 404 (Not Found).
    """
    logger = logging.getLogger(__name__)
    if data is not None:
        wstl_builder = request[requestproperty.HEA_WSTL_BUILDER]
        wstl_builder.href = str(request.url)
        try:
            data_ = data if isinstance(data, Sequence) else [data]
            body, mime_type = await wstl_builder.represent_from_request(request, data_,
                                                                        permissions=permissions,
                                                                        attribute_permissions=attribute_permissions,
                                                                        include_data=include_data)
            return _to_response(request, body, mime_type)
        except ValueError as e:
            logger.exception('Invalid input data %s', wstl_builder)
            return status_not_acceptable(str(e))
    else:
        return status_not_found()


def _to_response(request: web.Request,
                       body: bytes,
                       mime_type: str,
                       caching_strategy: CachingStrategy | None = None) -> web.Response:
    """
    Handle a get or get all request that returns one or more run-time WeSTL documents. Any actions in the documents are
    added to the request's run-time WeSTL documents, and the href of the action is prepended by the service's base URL.
    The actions in the provided run-time documents are expected to have a relative href.

    :param request: the HTTP request object. Cannot be None.
    :param run_time_docs: a list of run-time WeSTL documents containing data
    :param caching_strategy: what caching headers to send. By default, headers are generated to prevent caching.
    :return: aiohttp.web.Response object, with a 200 status code and the requested JSON objects in the body,
    status code 406 (Not Acceptable) if content negotiation failed to determine an acceptable content type, or status
    code 404 (Not Found).
    """
    _logger = logging.getLogger(__name__)
    _logger.debug('Run-time WeSTL document is %s', bytes)
    if caching_strategy:
        etag = _compute_etag(body)
        if request.if_none_match and etag in request.if_none_match:
            return web.HTTPNotModified()
    response = status_ok(body=body, content_type=mime_type, headers=NO_CACHE_HEADERS if not caching_strategy else None)
    if caching_strategy:
        response.etag = etag
    return response


async def _handle_get_result_from_wstl(request: web.Request,
                                       run_time_docs: Iterable[dict[str, Any]] | None,
                                       caching_strategy: CachingStrategy | None = None) -> web.Response:
    """
    Handle a get or get all request that returns one or more run-time WeSTL documents. Any actions in the documents are
    added to the request's run-time WeSTL documents, and the href of the action is prepended by the service's base URL.
    The actions in the provided run-time documents are expected to have a relative href.

    :param request: the HTTP request object. Cannot be None.
    :param run_time_docs: a list of run-time WeSTL documents containing data
    :param caching_strategy: what caching headers to send. By default, headers are generated to prevent caching.
    :return: aiohttp.web.Response object, with a 200 status code and the requested JSON objects in the body,
    status code 406 (Not Acceptable) if content negotiation failed to determine an acceptable content type, or status
    code 404 (Not Found).
    """
    _logger = logging.getLogger(__name__)
    _logger.debug('Run-time WeSTL document is %s', run_time_docs)
    if run_time_docs is not None:
        try:
            body, mime_type = await formats_from_request(request, run_time_docs)
            if caching_strategy:
                etag = _compute_etag(body)
                if request.if_none_match and etag in request.if_none_match:
                    return web.HTTPNotModified()
            response = status_ok(body=body, content_type=mime_type, headers=NO_CACHE_HEADERS if not caching_strategy else None)
            if caching_strategy:
                response.etag = etag
            return response
        except ValueError as e:
            return status_not_acceptable(str(e))
    else:
        return status_not_found()


async def _handle_post_result(request: web.Request, resource_base: str, inserted_id: str) -> web.Response:
    """
    Handle the result from a post request.

    :param request: the HTTP request object (required).
    :param resource_base: the common base path fragment for all resources of this type (required).
    :param inserted_id: the id of the newly created object (required).
    :return: aiohttp.web.Response object with status code 201 (Created).
    """
    return status_created(request.app[appproperty.HEA_COMPONENT], resource_base, inserted_id)


async def _get_options(methods: list[str]) -> web.Response:
    """
    Create and return a Response object in response to an OPTIONS request.

    :param methods: the allowed HTTP methods.
    :return: an aiohttp.web.Response object with a 200 status code and an Allow header.
    """
    resp = web.HTTPOk()
    resp.headers[hdrs.ALLOW] = ', '.join(methods)
    return resp


def _compute_etag(body: bytes | str, encoding='utf-8') -> ETag:
    """
    Computes an ETag from a bytes or str object.

    @param body: a bytes or str object.
    @param encoding: for bytes bodies, an optional encoding.
    @return: an ETag object.
    """
    if isinstance(body, str):
        body_ = body.encode(encoding)
    else:
        body_ = bytes(body)
    return ETag(md5(body_).hexdigest())


def get_async_status(request: web.Request) -> web.Response:
    """
    Implements asynchronous API status checking.

    :param request: the HTTP request (required).
    :return: the HTTP response.
    """
    logger = logging.getLogger(__name__)
    sub = request.headers.get(SUB, NONE_USER)
    background_tasks = request.app[HEA_BACKGROUND_TASKS]
    logger.debug('background tasks: %s', background_tasks)
    task_name = f'{sub}^{request.url}'
    if not background_tasks.contains(task_name):
        return status_not_found()
    if not background_tasks.done(task_name):
        return status_accepted()
    error = background_tasks.error(task_name)
    if error:
        background_tasks.remove(task_name)
        # In case we get an exception other than an HTTPException, raise it so it gets wrapped in an internal server
        # error response.
        raise error
    else:
        resp = background_tasks.result(task_name)
        background_tasks.remove(task_name)
        return resp



"""
Header constants for the Open ID Connect standard claims, and additional claims specific to HEA, that are set by the
HEA reverse proxy. The standard claim headers are prefixed by OIDC_CLAIM_.

Definitions of the standard claims may be found at
https://openid.net/specs/openid-connect-core-1_0.html#StandardClaims. This document also specifies requirements
for using additional claims, which HEA follows.

The following constants contain the names of the claim headers:
SUB: the sub claim.
AUD: the aud (audience) claim.
CLAIM_HEADERS: a tuple containing all of the claim header names.
"""

SUB = 'OIDC_CLAIM_sub'  # currently logged in user ("subject")
AUD = 'OIDC_CLAIM_aud'  # intended user
AZP = 'OIDC_CLAIM_azp'  # client id
ISS = 'OIDC_CLAIM_iss'  # OIDC provider URL

CLAIM_HEADERS = (SUB, AUD, AZP, ISS)



from unittest import IsolatedAsyncioTestCase
from heaserver.service import activity
from heaserver.service.util import yield_control
from aiohttp.test_utils import make_mocked_request
from aiohttp.web import Application, Request
from heaobject.root import DesktopObject
from heaobject.activity import Status
from heaserver.service import appproperty
from mypy_boto3_s3 import S3Client
from typing import Any
import asyncio

class ActivityTestCase(IsolatedAsyncioTestCase):

    async def test_desktop_object_actions_number(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        desktop_objects = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            desktop_objects.append(desktop_object)
        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual(2, len(desktop_objects))

    async def test_desktop_object_actions_number_nothing_awaited(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        desktop_objects = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            desktop_objects.append(desktop_object)
        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            pass

        self.assertEqual(1, len(desktop_objects))

    async def test_desktop_object_actions_status_succeeded(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        statuses = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            statuses.append(desktop_object.status)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual([Status.IN_PROGRESS, Status.SUCCEEDED], statuses)

    async def test_desktop_object_actions_status_code(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        codes = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            codes.append(desktop_object.code)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual(['hea-get', 'hea-get'], codes)

    async def test_desktop_object_actions_status_description(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        codes = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            codes.append(desktop_object.description)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual(['hello world', 'hello world'], codes)

    async def test_desktop_object_actions_status_began(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        timestamps = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            timestamps.append(desktop_object.began)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await asyncio.sleep(0.1)

        self.assertEqual([True, True], [began is not None for began in timestamps])

    async def test_desktop_object_actions_status_failed_began(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        timestamps = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            timestamps.append(desktop_object.began)

        with self.assertRaises(ValueError):
            async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
                await yield_control()
                raise ValueError

        self.assertEqual([True, True], [began is not None for began in timestamps])

    async def test_desktop_object_actions_status_failed_exception(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        timestamps = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            timestamps.append(desktop_object.ended)

        with self.assertRaises(ValueError):
            async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
                raise ValueError

    async def test_desktop_object_actions_status_failed_ended(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        timestamps = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            timestamps.append(desktop_object.ended)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual([False, True], [ended is not None for ended in timestamps])

    async def test_desktop_object_actions_status_ended(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        timestamps = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            timestamps.append(desktop_object.ended)

        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
            await yield_control()

        self.assertEqual([False, True], [ended is not None for ended in timestamps])

    async def test_desktop_object_actions_status_failed(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        statuses = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            statuses.append(desktop_object.status)

        with self.assertRaises(ValueError):
            async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as _:
                raise ValueError

    async def test_desktop_object_actions_status_failed_statuses(self):
        app = Application()
        request = make_mocked_request('GET', '/helloworld', app=app)
        app[appproperty.HEA_DB] = self.__mock_s3()
        statuses = []
        async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
            statuses.append(desktop_object.status)
        async with activity.DesktopObjectActionLifecycle(request, 'hea-get', 'hello world', activity_cb=publish_desktop_object) as activity_:
            await yield_control()
            activity_.status = Status.FAILED
        self.assertEqual([Status.IN_PROGRESS, Status.FAILED], statuses)

    def __mock_s3(self) -> Any:
        class MockS3:
            async def get_client(self, request: Request, service_name: str, volume_id: str) -> S3Client:
                return None
        return MockS3()





from unittest import TestCase
from heaserver.service.db import awsservicelib
from botocore.exceptions import ClientError

import heaserver.service.db.aws


class TestAWSServiceLib(TestCase):

    def test_handle_client_error_404(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_404}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_no_such_bucket(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_NO_SUCH_BUCKET}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_unknown(self):
        c = ClientError(error_response={'Error': {'Code': "It's wrecked"}},
                        operation_name='foo')
        self.assertEqual(500, awsservicelib.handle_client_error(c).status)




import logging
from typing import AsyncGenerator

from aiohttp import web
from aiohttp.web_request import Request
from aiohttp.web_response import Response
from heaobject.folder import Item

import re

from heaserver.service import response
from heaserver.service.appproperty import HEA_DB
from heaserver.service.db.database import DatabaseContextManager
from heaobject.root import DesktopObjectDict
from heaserver.service.db.opensearch import OpenSearch, ItemTypeVar, OpenSearchContext
from heaserver.service.oidcclaimhdrs import SUB


def build_query(request: Request, permission_context: dict[str, list[str]], regexp_type: str, search_term: str):
    # Placeholder for the must array in the query
    # Handle multiple `regexp` query params
    santize_search_term = re.escape(search_term).lower()
    if not search_term:
        raise ValueError("search text is required and cannot be empty")

    # Build the final query object
    search_path = None
    if regexp_type == 'contains':
        search_path= {
            "match": {
                "path.contains": santize_search_term
            }
        }
    elif regexp_type == 'starts_with':
        search_path= {
            "prefix": {
                "path.starts_with": santize_search_term
            }
        }
    elif regexp_type == 'ends_with':
        search_path= {
            "prefix": {
                "path.ends_with": santize_search_term[::-1]
            }
        }
    else:
        raise ValueError("Invalid query type. Must be one of: 'contains', 'starts_with', 'ends_with'.")

    query = {
        "query": {
            "bool": {
                "must": [
                    search_path,
                    {"terms": permission_context}
                ]
            }
        }
    }
    if regexp_type == 'contains':
        query["query"]["bool"]["filter"] = [
            {
                "regexp": {
                    "path.keyword": f".*{santize_search_term}.*"
                }
            }
        ]

    return query



async def insert (request: Request,
                  doc: Item,
                  index: str | None = None) -> Response:
    """
    Executes a dynamic search query in OpenSearch and returns the result.

    :param doc:
    :param request: the HTTP request.
    :param index: (Optional) the OpenSearch index name, if not present it will default to db config for it.
    :return: The OpenSearch query result or None if no result is found.
    """
    logger = logging.getLogger(__name__)
    sub = request.headers.get('SUB', 'none_user')

    # Build the query dynamically based on the input parameters
    async with OpenSearchContext(request=request, volume_id= doc.volume_id if doc.volume_id else None) as opensearch:
        result = await opensearch.insert(doc=doc.to_dict() , index=index)
        if not result:
            return response.status_bad_request("Invalid search results")
        logger.debug(f"opensearch result id: {result}")
    return await response.put(True if result else False)


async def batch_insert(request: Request,
                       doc_gen: AsyncGenerator[ItemTypeVar, None],
                       volume_id: str | None = None) -> Response:
    """
    Executes a dynamic search query in OpenSearch and returns the result.

    :param doc_gen: a generator that yields ItemTypeVar
    :param volume_id: (Optional) the volume if for associating the volume.
    :return: The OpenSearch query result or None if no result is found.
    """
    logger = logging.getLogger(__name__)

    # Build the query dynamically based on the input parameters
    async with OpenSearchContext(request=request, volume_id= volume_id) as opensearch:
        result = await opensearch.batch_insert(bulk_doc_gen=doc_gen)
        if not result:
            return response.status_bad_request("Invalid search results")
        logger.debug(f"opensearch result id: {result}")
    return await response.put(True if result else False)


async def search(request: Request,
                 search_item_type: type[ItemTypeVar],
                 perm_context: dict[str, list[str]],
                 index: str | None = None,
                 volume_id: str | None = None) -> Response:
    """
    Executes a dynamic search query in OpenSearch and returns the result.

    :param request: the HTTP request.
    :param search_item_type:
    :param perm_context: The permission context provides list of strings where at least one needs match
    :param index: (Optional) the OpenSearch index name, if not present it will default to db config for it.
    :param volume_id: (Optional) the id of the volume
    :return: The OpenSearch query result or None if no result is found.
    """
    sub = request.headers.get('SUB', 'none_user')
    logger = logging.getLogger(__name__)
    volume_id_ = request.match_info.get('volume_id') if request.match_info.get('volume_id', None) else volume_id
    regexp_type = request.query.get('regexp_type', 'contains')
    search_term = request.query.get('text', '')

    # Build the query dynamically based on the input parameters
    query = build_query(request, perm_context,regexp_type, search_term )
    async with OpenSearchContext(request=request, volume_id=volume_id_) as opensearch:
        results = await opensearch.search(query=query, search_item_type=search_item_type, index=index)
        if not results:
            return response.status_bad_request("Invalid search results")
        logger.debug(f"opensearch result: {results}")
    return await response.get_all(request, [sr.to_dict() for sr in results])


async def search_dict(request: Request,
                      search_item_type: type[ItemTypeVar],
                      perm_context: dict[str, list[str]],
                      index: str | None = None,
                      volume_id: str | None = None)  -> list[DesktopObjectDict]:
    """
    Executes a dynamic search query in OpenSearch and returns the results as a list of dictionaries.

    :param request: The HTTP request containing headers and query parameters.
    :param search_item_type: The type of items to search for, used for deserializing the results.
    :param perm_context: A dictionary of permission context specifying lists of strings where at least one needs to match.
    :param index: (Optional) The OpenSearch index name. Defaults to the configured database index if not specified.
    :param volume_id: (Optional) The ID of the volume to filter the search query.
    :return: A list of search results as dictionaries. Returns an empty list if no results are found.
    """
    logger = logging.getLogger(__name__)
    volume_id_ = request.match_info.get('volume_id') if request.match_info.get('volume_id', None) else volume_id
    regexp_type = request.query.get('regexp_type', 'contains')
    search_term = request.query.get('text', '')

    # Build the query dynamically based on the input parameters
    query = build_query(request, perm_context, regexp_type, search_term)
    async with OpenSearchContext(request=request, volume_id=volume_id_) as opensearch:
        results = await opensearch.search(query=query, search_item_type=search_item_type, index=index)
        if not results:
            return []
        logger.debug(f"opensearch result: {results}")
    return [sr.to_dict() for sr in results]

